@@ -6,13 +6,15 @@ from functools import partial
 import builtins
 from tx.functional.either import Left, Right, Either
 from tx.functional.maybe import Just, Nothing
-from .dependentqueue import DependentQueue, ResultType, ReturnType, RR
+from .dependentqueue import DependentQueue, ResultType, ReturnType
 from .utils import mappend
 from .spec import AbsSpec, LetSpec, MapSpec, CondSpec, PythonSpec, SeqSpec, RetSpec, TopSpec, AbsValue, NameValue, DataValue, ret_prefix_to_str, free_names, bound_names, sort_tasks, preproc_tasks
 from tx.readable_log import format_message, getLogger
-from typing import List, Any, Dict, Tuple, Set, Callable, TypeVar, ClassVar, Union
+from typing import List, Any, Dict, Tuple, Set, Callable, TypeVar, ClassVar, Union, TextIO
 from dataclasses import dataclass
 from abc import ABC, abstractmethod
+from .serialization import jsonify
+import json
 import pickle
 
 logger = getLogger(__name__, logging.INFO)
@@ -33,24 +35,30 @@ class AbsTask(ABC):
 class IdentifiedTask(AbsTask):
     task_id: str
     
+
+def write_output(output, obj):
+    objjsonified = jsonify(obj)
+    logger.info(format_message("write_output", "jsonified object to", {"obj jsonified as": objjsonified}))
+    output.write(json.dumps(objjsonified) + "\n")
+
 @dataclass
 class BaseTask(IdentifiedTask):
     log_error: ClassVar[bool] = False
     
-    def run(self, results: ReturnType, subnode_results: ReturnType, queue: DependentQueue) -> RR:
+    def run(self, results: ReturnType, subnode_results: ReturnType, queue: DependentQueue, output: TextIO) -> ResultType:
         logger.debug(format_message("BaseTask.run", "start", {"results": results}))
         try:
-            return mbind(self.baseRun, results, subnode_results, queue, log_error=self.log_error)
+            return mbind(self.baseRun, results, subnode_results, queue, output, log_error=self.log_error)
         except Exception as e:
             err = (str(e), traceback.format_exc())
             logger.error(str(err))
-            queue.put_output({":error:": Right(err)})
+            write_output(output, {":error:": Right(err)})
             queue.close()
             raise
             
 
     @abstractmethod
-    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue) -> RR:
+    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue, output: TextIO) -> ResultType:
         pass
 
 
@@ -64,7 +72,7 @@ class Task(BaseTask):
     args: Dict[int, Any]
     kwargs: Dict[str, Any]
 
-    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue) -> RR:
+    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue, output: TextIO) -> ResultType:
         try:
             logger.debug(format_message("Task.baseRun", "start", {"results": results}))
             mod = import_module(self.mod) if self.mod != "" else builtins
@@ -80,7 +88,7 @@ class Task(BaseTask):
             err = (str(e), traceback.format_exc())
             logger.error(str(err))
             result = Left(err)
-        return {}, Right({self.name: result})
+        return Right({self.name: result})
 
 
 @dataclass
@@ -102,7 +110,7 @@ class DynamicMap(BaseTask):
     level: int
     log_error: ClassVar[bool] = True
 
-    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue) -> RR:
+    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue, output: TextIO) -> ResultType:
         hold_id = queue.put(Hold(), is_hold=True)
         logger.debug("DynamicMap.baseRun: put hold task on queue %s", hold_id)
         logger.debug(format_message("DynamicMap.baseRun", "enqueue call", {"results": results, "results[self.coll_name]": results[self.coll_name]}))
@@ -116,9 +124,9 @@ class DynamicMap(BaseTask):
             hold={hold_id},
             level=self.level
         )
-        queue.complete(hold_id, {}, Right({}))
+        queue.complete(hold_id, Right({}))
         logger.debug("DynamicMap.baseRun: remove hold task from queue %s", hold_id)
-        return {}, Right({})
+        return Right({})
 
 
 @dataclass
@@ -131,7 +139,7 @@ class DynamicGuard(BaseTask):
     level: int
     log_error: ClassVar[bool] = True
     
-    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue) -> RR:
+    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue, output: TextIO) -> ResultType:
         hold_id = queue.put(Hold(), is_hold=True)
         logger.debug("DynamicCond.baseRun: put hold task on queue %s", hold_id)
         enqueue(
@@ -144,9 +152,9 @@ class DynamicGuard(BaseTask):
             hold={hold_id},
             level=self.level
         )
-        queue.complete(hold_id, {}, Right({}))
+        queue.complete(hold_id, Right({}))
         logger.debug("DynamicCond.baseRun: remove hold task from queue %s", hold_id)
-        return {}, Right({})
+        return Right({})
     
 
 @dataclass
@@ -154,8 +162,8 @@ class DynamicLet(BaseTask):
     name: str
     obj_name: str
     
-    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue) -> RR:
-        return {}, Right({self.name: results[self.obj_name]})
+    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue, output: TextIO) -> ResultType:
+        return Right({self.name: results[self.obj_name]})
 
 
 @dataclass
@@ -163,8 +171,8 @@ class Let(BaseTask):
     name: str
     obj: Either[Any, Any]
 
-    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue) -> RR:
-        return {}, Right({self.name: self.obj})
+    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue, output: TextIO) -> ResultType:
+        return Right({self.name: self.obj})
 
     
 @dataclass
@@ -172,8 +180,9 @@ class DynamicRet(IdentifiedTask):
     obj_name: str
     ret_prefix: List[Any]
     
-    def run(self, results: ReturnType, subnode_results: ReturnType, queue: DependentQueue) -> RR:
-        return {ret_prefix_to_str(self.ret_prefix): results[self.obj_name]}, Right({})
+    def run(self, results: ReturnType, subnode_results: ReturnType, queue: DependentQueue, output: TextIO) -> ResultType:
+        write_output(output, {ret_prefix_to_str(self.ret_prefix): results[self.obj_name]})
+        return Right({})
 
 
 @dataclass
@@ -181,8 +190,9 @@ class Ret(IdentifiedTask):
     obj: Either[Any, Any]
     ret_prefix: List[Any]
 
-    def run(self, results: ReturnType, subnode_results: ReturnType, queue: DependentQueue) -> RR:
-        return {ret_prefix_to_str(self.ret_prefix): self.obj}, Right({})
+    def run(self, results: ReturnType, subnode_results: ReturnType, queue: DependentQueue, output: TextIO) -> ResultType:
+        write_output(output, {ret_prefix_to_str(self.ret_prefix): self.obj})
+        return Right({})
 
     
 @dataclass
@@ -193,9 +203,9 @@ class Seq(BaseTask):
     task_id: str
     log_error: ClassVar[bool] = True
 
-    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue) -> RR:
+    def baseRun(self, results: Dict[str, Any], subnode_results: Dict[str, Any], queue: DependentQueue, output: TextIO) -> ResultType:
         data = {**self.data, **{name: Right(value) for name, value in results.items()}}
-        return evaluate(self.spec, data, self.ret_prefix)
+        return evaluate(self.spec, data, self.ret_prefix, output)
 
     
 @dataclass
@@ -219,64 +229,60 @@ def evaluate_value(data: ReturnType, arg: AbsValue) -> Either[Any, Any]:
         return Left(f"unsupported value {arg}")
 
 
-def evaluate(spec: AbsSpec, data: ReturnType, ret_prefix: List[Any]) -> RR:
+def evaluate(spec: AbsSpec, data: ReturnType, ret_prefix: List[Any], output: TextIO) -> ResultType:
     logger.debug(format_message("evaluate", "executing sequentially", {"spec": spec, "data": data, "ret_prefix": ret_prefix}))
     if isinstance(spec, LetSpec):
         name = spec.name
         obj = evaluate_value(data, spec.obj)
-        return {}, obj.bind(lambda _: {name: obj})
+        return obj.bind(lambda _: {name: obj})
     elif isinstance(spec, MapSpec):
         coll_value = spec.coll
         var = spec.var
         subspec = spec.sub
         coll = evaluate_value(data, coll_value)
         if isinstance(coll, Left):
-            return {":error:": Right(coll.value)}, Right({})
+            write_output(output, {":error:": Right(coll.value)})
+            return Right({})
         coll = coll.value
-        ret = {}
         for i, row in enumerate(coll):
             data2 = {**data, var:Right(row)}
-            sub_ret, sub_result = evaluate(subspec, data2, ret_prefix=ret_prefix + [i])
-            ret.update(sub_ret)
+            sub_result = evaluate(subspec, data2, ret_prefix=ret_prefix + [i], output=output)
             if isinstance(sub_result, Left):
-                return ret, sub_result
-        return ret, Right({}) # ignore all sub_results
+                return sub_result
+        return Right({}) # ignore all sub_results
     elif isinstance(spec, CondSpec):
         cond_value = spec.on
         then_spec = spec.then
         else_spec = spec._else
         cond = evaluate_value(data, cond_value)
         if isinstance(cond, Left):
-            return {":error:": Right(cond.value)}, Right({})
+            write_output(output, {":error:": Right(cond.value)})
+            return Right({})
         cond = cond.value
         if cond:
-            return evaluate(then_spec, data, ret_prefix=ret_prefix)
+            return evaluate(then_spec, data, ret_prefix=ret_prefix, output=output)
         else:
-            return evaluate(else_spec, data, ret_prefix=ret_prefix)
+            return evaluate(else_spec, data, ret_prefix=ret_prefix, output=output)
     elif isinstance(spec, TopSpec):
         subs = spec.sub
         subs_sorted = sort_tasks(set(data.keys()), subs)
-        ret = {}
         for sub in subs_sorted:
-            sub_ret, sub_result = evaluate(sub, data, ret_prefix=ret_prefix)
-            ret = mappend(ret, sub_ret)
+            sub_result = evaluate(sub, data, ret_prefix=ret_prefix, output=output)
             if isinstance(sub_result, Left):
-                return ret, sub_result
+                return sub_result
             data = {**data, **sub_result.value}
-        return ret, Right({})
+        return Right({})
     elif isinstance(spec, SeqSpec):
         subs = spec.sub
         subs_sorted = sort_tasks(set(data.keys()), subs)
-        ret = {}
         result : ResultType = {}
         for sub in subs_sorted:
-            sub_ret, sub_result = evaluate(sub, data, ret_prefix=ret_prefix)
-            ret = mappend(ret, sub_ret)
+            sub_result = evaluate(sub, data, ret_prefix=ret_prefix, output=output)
             if isinstance(sub_result, Left):
-                return ret, sub_result
+                return sub_result
             data = {**data, **sub_result.value}
             result.update(sub_result.value)
-        return ret, Right(result)
+        return Right(result)
     elif isinstance(spec, PythonSpec):
         try:
             mod = import_module(spec.mod) if spec.mod != "" else builtins
@@ -286,7 +292,7 @@ def evaluate(spec: AbsSpec, data: ReturnType, ret_prefix: List[Any]) -> RR:
             args0 = {k: evaluate_value(data, v) for k, v in spec.params.items()}
             errors = [x for x in args0.values() if isinstance(x, Left)]
             if len(errors) > 0:
-                return {}, Right({name: errors[0]})
+                return Right({name: errors[0]})
             args2 = {k: v.value for k, v in args0.items()}
             args, kwargs = split_args(args2)
             logger.debug(format_message("evaluate", "PythonSpec before", {"spec.name": spec.name, "spec.mod": spec.mod, "func": lambda: func, "spec.func" : spec.func, "spec.params": spec.params, "args": args, "kwargs": kwargs}))
@@ -298,24 +304,24 @@ def evaluate(spec: AbsSpec, data: ReturnType, ret_prefix: List[Any]) -> RR:
         except Exception as e:
             result = Left((str(e), traceback.format_exc()))
         logger.debug(format_message("evaluate", "PythonSpec", {"result": result}))
-        return {}, Right({spec.name: result})
+        return Right({spec.name: result})
     elif isinstance(spec, RetSpec):
         obj_value = spec.obj
         obj = evaluate_value(data, obj_value)
-        return {ret_prefix_to_str(ret_prefix): obj}, Right({})
+        write_output(output, {ret_prefix_to_str(ret_prefix): obj})
+        return Right({})
     else:
         raise RuntimeError(f'unsupported spec type {spec}')
     
     
-def mbind(job_run : Callable[[Dict[str, Any], Dict[str, Any], DependentQueue], RR], params: ReturnType, subnode_results: ReturnType, queue: DependentQueue, log_error: bool) -> RR:
+def mbind(job_run : Callable[[Dict[str, Any], Dict[str, Any], DependentQueue, TextIO], ResultType], params: ReturnType, subnode_results: ReturnType, queue: DependentQueue, output: TextIO, log_error: bool) -> ResultType:
     resultv = {}
     subnode_resultv = {}
     for k, v in params.items():
         if isinstance(v, Left):
             resultj = v
-            ret: Dict[str, Any] = {}
             if log_error:
-                queue.put_output({":error:": Right(v.value)})
+                write_output(output, {":error:": Right(v.value)})
             break
         else:
             resultv[k] = v.value
@@ -323,16 +329,15 @@ def mbind(job_run : Callable[[Dict[str, Any], Dict[str, Any], DependentQueue], R
         for k, v in subnode_results.items():
             if isinstance(v, Left):
                 resultj = v
-                ret = {}
                 if log_error:
-                    queue.put_output({":error:": Right(v.value)})
+                    write_output(output, {":error:": Right(v.value)})
                 break
             else:
                 subnode_resultv[k] = v.value
         else:
             # logger.debug(f"mbind: running {job_run}")
-            ret, resultj = job_run(resultv, subnode_resultv, queue)
-    return ret, resultj
+            resultj = job_run(resultv, subnode_resultv, queue, output)
+    return resultj
 
                         
 def split_args(args0: Dict[Union[int, str], Any]) -> Tuple[Dict[int, Any], Dict[str, Any]]:
@@ -373,14 +378,16 @@ def generate_tasks(queue: DependentQueue, spec: AbsSpec, data: ReturnType, env:
         var = spec.var
         subspec = spec.sub
         subnode_ret_prefix = ret_prefix + ["@map"]
+        free_names_sub_with_possibly_var = free_names(subspec)
+        var_in_sub = var in free_names_sub_with_possibly_var
+        free_names_sub = free_names_sub_with_possibly_var - {var}
+        subnode_env = get_submap(env, free_names_sub)
+        subnode_data = get_submap(data, free_names_sub)
         if isinstance(coll_value, NameValue) and (coll_name := coll_value.name) not in data:
             coll_source = env[coll_name]
-            free_names_sub = free_names(subspec) - {var}
-            subnode_env = get_submap(env, free_names_sub)
-            subnode_data = get_submap(data, free_names_sub)
             task: IdentifiedTask = DynamicMap(ret_prefix_to_str(subnode_ret_prefix, False), var, coll_name, subspec, subnode_data, ret_prefix, level)
             dep = {coll_source: {coll_name}}
-            enqueue_task(queue, task, {**dep, **hold_dep}, inverse_dict(subnode_env))
+            enqueue_task(queue, task, {**dep, **hold_dep}, inverse_dict(subnode_env), set())
         else:
             coll = evaluate_value(data, coll_value)
 
@@ -390,20 +397,17 @@ def generate_tasks(queue: DependentQueue, spec: AbsSpec, data: ReturnType, env:
             logger.debug(format_message("generate_tasks", "evaluate_value call ret val", {"data": data, "coll_value": coll_value, "coll": coll}))
             
             for i, row in enumerate(coll.value):
-                data_sub = {**data, var: Right(row)}
+                subnode_data_with_possibly_var = {**subnode_data, **({var: Right(row)} if var_in_sub else {})}
                 subnode_ret_prefix_i = subnode_ret_prefix + [i]
                 if level > 0:
-                    generate_tasks(queue, subspec, data=data_sub, env=env, ret_prefix=subnode_ret_prefix_i, hold=hold, level=level-1)
+                    generate_tasks(queue, subspec, data=subnode_data_with_possibly_var, env=env, ret_prefix=subnode_ret_prefix_i, hold=hold, level=level-1)
                 else:
-                    free_names_sub = free_names(subspec)
-                    subnode_env = get_submap(env, free_names_sub - {var})
-                    subnode_data = get_submap(data_sub, free_names_sub)
-                    task = Seq(ret_prefix_to_str(subnode_ret_prefix_i, False), subspec, subnode_data, subnode_ret_prefix_i)
+                    task = Seq(ret_prefix_to_str(subnode_ret_prefix_i, False), subspec, subnode_data_with_possibly_var, subnode_ret_prefix_i)
                     logger.info(format_message("generate_tasks", "generating Seq task", {
                         "id": ret_prefix_to_str(subnode_ret_prefix_i, False),
                         "data": lambda: dict_size(subnode_data)
                     }))
-                    enqueue_task(queue, task, {**inverse_dict(subnode_env), **hold_dep}, {})
+                    enqueue_task(queue, task, {**inverse_dict(subnode_env), **hold_dep}, {}, set())
 
     elif isinstance(spec, CondSpec):
         cond_value = spec.on
@@ -417,7 +421,7 @@ def generate_tasks(queue: DependentQueue, spec: AbsSpec, data: ReturnType, env:
             subnode_data = get_submap(data, free_names_sub)
             task = DynamicGuard(ret_prefix_to_str(subnode_ret_prefix, False), cond_name, then_spec, else_spec, subnode_data, ret_prefix, level)
             dep = {cond_source: {cond_name}}
-            enqueue_task(queue, task, {**dep, **hold_dep}, inverse_dict(subnode_env))
+            enqueue_task(queue, task, {**dep, **hold_dep}, inverse_dict(subnode_env), set())
         else:
             cond = evaluate_value(data, cond_value)
 
@@ -442,6 +446,7 @@ def generate_tasks(queue: DependentQueue, spec: AbsSpec, data: ReturnType, env:
             generate_tasks(queue, sub, data=data, env=env_sub, ret_prefix=subnode_ret_prefix + [task_name], hold=hold, level=level)
     elif isinstance(spec, SeqSpec):
         free_names_sub = free_names(spec)
+        bound_names_sub = bound_names(spec)
         env_sub = get_submap(env, free_names_sub)
         data_sub = get_submap(data, free_names_sub)
         ret_prefix_sub = ret_prefix + ["@seq"]
@@ -450,7 +455,7 @@ def generate_tasks(queue: DependentQueue, spec: AbsSpec, data: ReturnType, env:
             "id": ret_prefix_to_str(ret_prefix, False),
             "data": lambda: dict_size(data_sub)
         }))
-        enqueue_task(queue, task, {**inverse_dict(env_sub), **hold_dep}, {})
+        enqueue_task(queue, task, {**inverse_dict(env_sub), **hold_dep}, {}, bound_names_sub)
     elif isinstance(spec, LetSpec):
         name = spec.name
         obj_value = spec.obj
@@ -458,11 +463,11 @@ def generate_tasks(queue: DependentQueue, spec: AbsSpec, data: ReturnType, env:
             obj_source = env[obj_name]
             task = DynamicLet(env[name], name, obj_name)
             dep = {obj_source: {obj_name}}
-            enqueue_task(queue, task, {**dep, **hold_dep}, {})
+            enqueue_task(queue, task, {**dep, **hold_dep}, {}, set())
         else:
             obj = evaluate_value(data, obj_value)
             task = Let(env[name], name, obj)
-            enqueue_task(queue, task, {}, {})
+            enqueue_task(queue, task, {}, {}, {name})
     elif isinstance(spec, PythonSpec):
         name = spec.name
         mod = spec.mod
@@ -481,7 +486,7 @@ def generate_tasks(queue: DependentQueue, spec: AbsSpec, data: ReturnType, env:
         free_names_sub = free_names(spec)
         env_sub = get_submap(env, free_names_sub)
         task = Task(env[name], name, mod, func, args_spec, kwargs_spec, args, kwargs)
-        enqueue_task(queue, task, {**inverse_dict(env_sub), **hold_dep}, {})
+        enqueue_task(queue, task, {**inverse_dict(env_sub), **hold_dep}, {}, {name})
         logger.debug(format_message("generate_tasks", "enqueue task", {"task": lambda: vars(task), "env[name]": lambda: env[name]})) 
     elif isinstance(spec, RetSpec):
         obj_value = spec.obj
@@ -490,21 +495,21 @@ def generate_tasks(queue: DependentQueue, spec: AbsSpec, data: ReturnType, env:
             obj_source = env[obj_name]
             task = DynamicRet(ret_prefix_to_str(subnode_ret_prefix, False), obj_name, subnode_ret_prefix)
             dep = {obj_source:{obj_name}}
-            enqueue_task(queue, task, {**dep, **hold_dep}, {})
+            enqueue_task(queue, task, {**dep, **hold_dep}, {}, set())
         else:
             obj = evaluate_value(data, obj_value)
             task = Ret(ret_prefix_to_str(subnode_ret_prefix, False), obj, subnode_ret_prefix)
-            enqueue_task(queue, task, hold_dep, {})
+            enqueue_task(queue, task, hold_dep, {}, set())
     else:
         raise RuntimeError(f'unsupported spec type {spec}')
 
 
-def enqueue_task(job_queue: DependentQueue, job: IdentifiedTask, depends_on: Dict[str, Set[str]], subnode_depends_on: Dict[str, Set[str]]) -> None:
+def enqueue_task(job_queue: DependentQueue, job: IdentifiedTask, depends_on: Dict[str, Set[str]], subnode_depends_on: Dict[str, Set[str]], names: Set[str]) -> None:
     
     logger.debug(format_message("enqueue_task", "start", {"input": job, "depends_on": depends_on, "subnode_depends_on": subnode_depends_on}))
     job_id = job.task_id
     logger.debug(format_message("enqueue_task", job_id, {"depends_on": depends_on, "subnode_depends_on": subnode_depends_on}))
-    job_queue.put(job, job_id=job_id, depends_on=depends_on, subnode_depends_on=subnode_depends_on)
+    job_queue.put(job, job_id=job_id, depends_on=depends_on, subnode_depends_on=subnode_depends_on, names=names)
 
     
 def enqueue(spec: AbsSpec, data: ReturnType, job_queue: DependentQueue, env: Dict[str, str]={}, ret_prefix: List[Any]=[], execute_original: bool=False, hold: Set[str]=set(), level:int=0) -> None:
